/*
 * Beyond_Arduino 5 - 1 Finger on Buzzer.c
 *
 *	Nuts & Volts - Beyond Arduino #5
 *  --------------------------------
 *	A game to test reaction time.
 *  2 LEDs (red and green) flash at a specific frequency
 *  Player 1 needs to press a button when the green LED is lit.
 *  WIN = green LED is lit / LOSE = red LED is lit.
 *  SCORE = lowest number of flashes before a WIN
 *  Player 2 controls the game over Serial (UART) connection:
 *   - "S" = start a new game
 *   - "1" to "9" = controls the LED flash speed
  *
 *  Connections:
 *    PB0 = red LED
 *    PB1 = green LED
 *    PD6 = push button
 *
 *  Author: Andrew Retallack
 *			www.crash-bang.com
 */ 


#define F_CPU 16000000UL	//Clock running at 16MHz. Need to define this prior to including "delay.h"


#include <avr/io.h>
#include <avr/interrupt.h>	//Include this file to help with interrupt handling
#include <util/delay.h>		//Library to handle delays


//Use macros for I/O pins for flexibility / ease of reading code
#define LED_RED PINB0
#define LED_GREEN PINB1
#define BUTTON PIND6

//Function Prototypes
void writeHelpUART(void);
void startSequence(void);
void processButtonPress(void);
void processUART(void);

void UART_Init(uint16_t baudRate);
void UART_writeChar(unsigned char data);
void UART_writeString(const char dataString[]);
void UART_printDecimal(uint16_t what);


//Variables
volatile uint8_t flashFreq = 9;		//Frequency at which LED flashes.  1 = fast / 9 = slow
volatile uint16_t  flashCount = 0;	//Count of number of flashes. The Player's score

volatile uint8_t uartReadByte;		//Byte read in over the UART
volatile uint8_t flagUART = 0;		//Interrupt Flag: 1 indicates data read in from the UART
volatile uint8_t flagButton = 0;	//Interrupt Flag: 1 indicates button was pressed


int main(void)
{
	uint8_t iCount;	//Counter

	//Configure I/O pins
	DDRB |= (1<<LED_RED) | (1<<LED_GREEN);			//Set LEDs as output
	PORTB &= ~( (1<<LED_RED) | (1<<LED_GREEN) );	//Ensure LEDs off
	
	DDRD &= ~( (1<<BUTTON) );	//Set Button as input
	PORTD |= (1<<BUTTON);		//Enable Pull-up resistor on Button

	
	UART_Init(9600);	//Initialise the UART at 9600 baud
	writeHelpUART();	//Write Help Text on UART
	
	//Initialise Interrupt: Pin Change
	PCICR |= (1<<PCIE2);	//Enable interrupts on PCINT "bank" 2: Button (PD6) is PCINT22, which is on "bank" 2 (therefore PCIE2)
	PCMSK2 |= (1<<PCINT22);	//Enable interrupts on pin PD6, which is PCINT22 (within "bank" 2)
	
	//Initialise Interrupt: UART Receive	
	//(Normally we would do this in the UART init function, but shown here for purpose of this example)
	UCSR0B |= (1<<RXCIE0);
	
	
	startSequence();	//Call the start sequence - flash LED's to show game beginning
	
	//Enable Global Interrupts
	sei();

    while(1)
    {
        
		flashCount++;	//Increase the counter for number of flashes (the score)
		
		//Toggle the LEDs
		PORTB ^= (1<<LED_RED);
		PORTB ^= (1<<LED_GREEN);
		
		//Delay between flashes.  Use the flashFreq received over UART. (Use a loop, as the "_delay_ms" needs a constant passed to it.)
		for (iCount = 0; iCount <= flashFreq; iCount++)
		{
			_delay_ms(50);
		}
		

		/*************************************************
		*	Handle flags returned by Interrupt Handlers
		*/
		
		if (flagButton == 1) //Button was pressed - the flagButton variable is set in the ISR
		{
			processButtonPress();
		}
		
		if (flagUART == 1)	//A character was received over the UART
		{
			processUART();
		}
		
		/*
		*	Finished handling interrupt flags
		*************************************************/

    }
}


/***********************************************************************
*
* Routine to write standard Help message over the UART
*
* INPUT:	none
* RETURNS:	none
************************************************************************/
void writeHelpUART(void)
{

	UART_writeString("Finger on the Buzzer\r\n");
	UART_writeString("====================\r\n");
	UART_writeString("Press 'S' to Start New Game\r\n");
	UART_writeString("Press '1' to '9' to change speed\r\n");
	UART_writeString("Goal: Press the button when the GREEN led is on\r\n");
	
}

/***********************************************************************
*
* Routine to show game is starting by flashing LEDs
*
* INPUT:	none
* RETURNS:	none
************************************************************************/
void startSequence(void)
{
	uint8_t iCount = 0;
	
	//Three flashes of both LEDs
	for (iCount=0; iCount < 3; iCount++)
	{
		PORTB |= (1<<LED_GREEN) | (1<<LED_RED);
		_delay_ms(500);
		PORTB &= ~ ( (1<<LED_GREEN) | (1<<LED_RED) );
		_delay_ms(250);
	}
	
	//Then both LEDs on for short period
	PORTB |= (1<<LED_RED) | (1<<LED_GREEN);
	_delay_ms(1000);
	
	//Then turn one LED off to make toggling in the main loop simpler
	PORTB &= ~(1<<LED_RED);
	
}


/***********************************************************************
*
* Routine to handle the processing once the button has been pressed:
*   - Check whether game won or lost
*   - Feedback to player and restart
*
* INPUT:	none
* RETURNS:	none
************************************************************************/
void processButtonPress(void)
{
	uint8_t iCount;
	
	//Disable interrupts - we need to finish this routine without interruption
	cli();
	
	//Check whether the Green LED is on (Datasheet says PORT register is read/write)
	if (PORTB & (1<<LED_GREEN) )  //Use bitmasking
	{
		//Message over UART
		UART_writeString("Got It!  You pressed when the LED was green.  Total flashes: ");
		UART_printDecimal(flashCount);
		UART_writeString("\r\n");

		//Flash Green LED 3 times
		for (iCount=0; iCount<7; iCount++)
		{
			PORTB ^= (1<<LED_GREEN);
			_delay_ms(500);
		}
				
		flashCount = 0;	//Reset score
				
	}
			
	else //RED LED must be on
	{
		//Message over UART
		UART_writeString("You Missed It!  You pressed when the LED was RED.  Total flashes so far: ");
		UART_printDecimal(flashCount);
		UART_writeString("\r\n");

		//Flash red LED 5 times
		for (iCount=0; iCount<11; iCount++)
		{
			PORTB ^= (1<<LED_RED);
			_delay_ms(250);
		}
				
	}
			
	startSequence(); //Start again
	
	flagButton = 0;	//Reset the flag for next button press
	

	//Check whether the user pressed the button and caused interrupt flag to set
	//(remember this flag can even be set when Global Interrupts are disabled)
	if (PCIFR & (1<<PCIF2))
	{
		 PCIFR |= (1<<PCIF2);	//Clear the Pin Change interrupt flag
	}
	
	//Re-enable interrupts
	PCICR |= (1<<PCIE2);	//Enable interrupts on PCINT "bank" 2
	sei();					//Enable global interrupts
	
}


/***********************************************************************
*
* Routine to handle the processing once a byte is received over UART:
*   - Reset a game
*   - Change flash speed
*
* INPUT:	none
* RETURNS:	none
************************************************************************/
void processUART(void)
{
	
	//Received "S": Restart the game
	if ( (uartReadByte == 'S') || (uartReadByte == 's') )
	{
		startSequence();
		flashCount = 0;	//Reset score
	}

	//Received from "1" to "9" (ASCII 49 to 57)
	if ( (uartReadByte >= 49) && (uartReadByte <= 57) )	
	{
		flashFreq = uartReadByte - 48;	//Convert ASCII to a number from 1 - 9 and set the Flash speed
	}
	
	flagUART = 0;	//Reset flag

}



/***********************************************************************
*
* This is the ISR (interrupt service routine) for PCINT2
*   It is triggered on a pin-change for "bank" 2
*
************************************************************************/
ISR(PCINT2_vect)
{
	PCICR &= ~(1<<PCIE2);	//Disable interrupts on PCINT "bank" 2: prevents button-bounce, or interrupt on switch release
	flagButton = 1;	//Set the flag so the main routine knows to handle UART event
}


/***********************************************************************
*
* This is the ISR (interrupt service routine) for UART_RX
*   It is triggered when a byte is received over the UART
*
************************************************************************/ISR(USART_RX_vect)
{
	uartReadByte = UDR0;	//Read the received byte from the UART
	flagUART = 1;	//Set the flag so the main routine knows to handle button press
}



/***********************************************************************
*
* This routine initialises the UART in Asynchronous mode by:
*   1. setting baud rate
*   2. setting data frame (8 bits, no parity, 1 stop bit)
*   3. enables the transmitter and received
*
* INPUT:	baudRate	The baud rate to operate at
* RETURNS:	none
************************************************************************/
void UART_Init(uint16_t baudRate)
{
	uint16_t my_UBRR;

	//Calculate the value of the UBRR register, to set the Baud Rate
	my_UBRR = (F_CPU/16UL/baudRate) - 1;

	//Set the Baud ratee
	UBRR0H = (unsigned char)(my_UBRR >> 8);		//Set the High register
	UBRR0L = (unsigned char)my_UBRR;			//Set the Low register

	//Stop Bit: 1 - ensure bit is unset (incase it was set before)
	UCSR0C &= ~(1<<USBS0);

	//Data Bits: 8
	UCSR0C |= ( (1<<UCSZ00) | (1<<UCSZ01) );

	//Parity: None - ensure bits unset (incase they were set before)
	UCSR0C &= ~( (1<<UPM00) | (1<<UPM01) );

	//Mode: Ensure Asynchronous Mode (clear bits incase they were set before)
	UCSR0C &= ~( (1<<UMSEL00) | (1<<UMSEL01) );
	
	//Enable the Transmitter and Receiver
	UCSR0B |= (1<<TXEN0) | (1<<RXEN0);

}


/***********************************************************************
*
* This routine writes a single character to the UART
*
* INPUT:	data	The character to write
* RETURNS:	none
************************************************************************/
void UART_writeChar(unsigned char data)
{
	
	//Wait until transmit buffer empty
	while ( !(UCSR0A & (1<<UDRE0) ) );
	
	//write char
	UDR0 = data;
	
}


/***********************************************************************
*
* This routine writes a string of characters to the UART
*
* INPUT:	dataString	The string to write
* RETURNS:	none
************************************************************************/
void UART_writeString(const char dataString[])
{
	uint16_t i = 0;
	
	//Loop through the string, writing character by character
	while(dataString[i])
	{
		UART_writeChar(dataString[i]);	//write each string
		i++;
	}
}


/***********************************************************************
*
* This routine writes a numeric value out in digits.  The value needs to
*	be converted to a string of ASCII characters.
*
* INPUT:	what	The value to write
* RETURNS:	none
************************************************************************/
void UART_printDecimal(uint16_t what)
{
	
	char numberText[6]="";
	uint16_t workingNumber = what;
	char tempWorkNum;
	uint16_t workingDivisor = 10000;
	uint8_t digitNumber = 0;
	
	//Generate the individual digits by successively dividing by multiples of 10, and storing in a string variable
	for (char i = 4; i>0; i--)
	{
		tempWorkNum = (workingNumber / (uint16_t)( workingDivisor ));
		
		//If the tempWorkNum is zero, then only record the digit if this is not the first digit (i.e. discard leading zeros)
		//If the tempWorkNum is non-zero, then record the digit
		if ( ((tempWorkNum == 0) && digitNumber > 0) || (tempWorkNum > 0) )
		{
			numberText[digitNumber] = tempWorkNum + 48;
			digitNumber++;
			
		}
		
		//Move one more digit to the right
		workingNumber = workingNumber % workingDivisor;	//Remove the digit already processed
		workingDivisor /= 10;
	}
	
	//The last digit in the number
	numberText[digitNumber] = (workingNumber % 10) + 48;
	
	//Finally, write the number out as text.
	UART_writeString(numberText);

}

